/*
 * pncc_fe - A speech recognition front-end for extracting Power
 * Nonrmalized Cepstral Coefficients
 *
 * Copyright (c) 2013 Emanuele Principi
 * Copyright (c) 2013 A3Lab, Department of Information Engineering,
 * Universita' Politecnica delle Marche, Ancona (ITALY).
 *
 * This file is part of pncc_fe.
 *
 * pncc_fe is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * pncc_fe is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with pncc_fe.  If not, see <http://www.gnu.org/licenses/>.
 *
 * List of changes:
 * v0.7.1: initial release.
 * v0.8.0: added support for HTK output format.
 */

#include <stdio.h>
#include <stdlib.h>
#include "a3_pncc.h"
#include <sndfile.h>
#include <string.h>
#include <sys/time.h>
#include <limits.h>
#include <getopt.h>
#include "iohtk.h"

#if defined(USE_DOUBLE)
#define PRECISION "double precision"
#define PRECISION_SUFF "double"
#define sf_read sf_read_double
#elif defined(USE_FLOAT)
#define PRECISION_SUFF "float"
#define PRECISION "single precision"
#define sf_read sf_read_float
#endif

#define SPHINX 0
#define HTK    1

#define VERSION "0.8.1"

void PrintUsage()
{
  printf("Usage: pncc_fe_%s [OPTIONS]\n", PRECISION_SUFF);
  printf("      -i, --input=FILEPATH     the input file path.\n");
  printf("      -o, --output=FILEPATH    the output file path.\n");
  printf("      -s, --sample-rate=VALUE  sample rate in Hz. One of 8000, 16000 (default).\n");
  printf("      -a, --start-time=VALUE   startTime in ms (default: 0).\n");
  printf("      -e, --end-time=VALUE     endTime in ms (default: end of file).\n");
  printf("      -c, --cmn                enables Cepstral Mean Normalization (default: no).\n");
  printf("      -n, --cmvn               enables Cepstral Mean and Variance Normalization (default: no).\n");
  printf("      -m, --mu                 initial value of the mean power estimate (default: see libpncc).\n");
  printf("      -f, --format             file output format. One of htk or sphinx (default).\n");
  printf("      -r, --raw                selects raw format for the input file. Default: raw if the file\n");
  printf("                               extension is '.raw'.\n");
  printf("      -t, --timing             enables printing of the signal duration, the processing time and\n");
  printf("                               the real-time factor.\n");
  printf("      -v, --version            prints out the current version and exits.\n");
  exit(EXIT_FAILURE);
}

void ProcessFile(char* input, char* output, int outformat, a3_pncc_t* pncc_state, bool printTiming, long int startTime,
                long int endTime, bool forceRaw)
{
  SNDFILE *fpInput;
  SF_INFO sfinfo;
  FILE* fpOutput;
  char* ch;
  sf_count_t nRead;
  int res;
  size_t nWritten;
  a3_real* samples, *pnccs;
  float* outbuf;
  int32_t nCoefs, i;
  struct timeval start, end, diff;
  sf_count_t startSample, endSample;
  HTKHeader htkHeader;

  printf("Configuration:\n");
  printf("  sample rate: %d\n", pncc_state->config->sampleRate);
  printf("  cmn: %s\n", pncc_state->config->cmnOffline ? "yes" : "no");
  printf("  cmvn: %s\n", pncc_state->config->cmvnOffline ? "yes" : "no");

  if (!forceRaw)
  {
    ch = strstr(input, ".raw");
    while (ch != NULL && strlen(ch) > 4)
      ch = strstr(ch+1, ".raw");
  }

  if (ch != NULL || forceRaw)
  {
    sfinfo.channels   = 1;
    sfinfo.samplerate = pncc_state->config->sampleRate;
    sfinfo.format     = SF_FORMAT_RAW | SF_ENDIAN_LITTLE | SF_FORMAT_PCM_16;
  }

   fpInput = sf_open(input, SFM_READ, &sfinfo);
  char* err = sf_strerror(fpInput);
  if (fpInput == NULL)
  {
    fprintf(stderr, "ERROR: unable to open '%s'\n", input);
    exit(EXIT_FAILURE);
  }
  else
  {
    sf_command(fpInput, SFC_SET_NORM_FLOAT, NULL, SF_FALSE);
    sf_command(fpInput, SFC_SET_NORM_DOUBLE, NULL, SF_FALSE);
    if (sfinfo.channels > 1)
    {
      fprintf(stderr, "ERROR: cannot process multichannel files (detected: %d).\n", sfinfo.channels);
      fprintf(stderr, "frames: %d\nsamplerate: %d\nchannels: %d\nformat: %d\nsections: %d\nseekable: %d\n",
              sfinfo.frames, sfinfo.samplerate, sfinfo.channels, sfinfo.format, sfinfo.sections, sfinfo.seekable);
      sf_close(fpInput);
      exit(EXIT_FAILURE);
    }

    if (sfinfo.samplerate != pncc_state->config->sampleRate)
    {
      fprintf(stderr, "ERROR: sample rate mismatch.\n");
      sf_close(fpInput);
      exit(EXIT_FAILURE);
    }

    startSample = 0;
    if (startTime != LONG_MIN)
    {
      startSample = (sf_count_t)floor(startTime / 1000.0 * sfinfo.samplerate);
      if (startSample > sfinfo.frames)
      {
        fprintf(stderr, "ERROR: start-time is greater than the signal duration.");
        sf_close(fpInput);
        exit(EXIT_FAILURE);
      }

      sf_seek(fpInput, startSample, SEEK_SET);
    }

    endSample = sfinfo.frames;
    if (endTime != LONG_MAX)
    {
      endSample = (sf_count_t)floor(endTime / 1000.0 * sfinfo.samplerate);
      if (endSample > sfinfo.frames)
      {
        fprintf(stderr, "ERROR: end-time is greater than the signal duration.");
        sf_close(fpInput);
        exit(EXIT_FAILURE);
      }
    }

    sf_count_t nSamples = endSample - startSample;

    samples = (a3_real*)malloc(sizeof(a3_real)*nSamples);
    if (samples == NULL)
    {
      fprintf(stderr, "ERROR: error allocating samples.\n");
      sf_close(fpInput);
      exit(EXIT_FAILURE);
    }

    nRead = sf_read(fpInput, samples, nSamples);
    sf_close(fpInput);
    if (nRead != nSamples)
    {
      fprintf(stderr, "ERROR: error reading samples.\n");
      exit(EXIT_FAILURE);
    }

    gettimeofday(&start, NULL);

    pnccs = a3_pncc_process_utt_alloc(samples, nRead, pncc_state);
    gettimeofday(&end, NULL);

    free(samples);
    if (pnccs == NULL)
    {
      fprintf(stderr, "ERROR: error processing samples.\n");
      exit(EXIT_FAILURE);
    }

    fpOutput = fopen(output, "wb");
    if (fpOutput == NULL)
    {
      fprintf(stderr, "ERROR: error opening the output file '%s'.\n", output);
      exit(EXIT_FAILURE);
    }

    nCoefs = a3_pncc_get_nframes(nSamples, pncc_state) * pncc_state->config->nCeps;
    if (outformat == HTK)
    {
      htkHeader.n_samples     = a3_pncc_get_nframes(nSamples, pncc_state);
      htkHeader.sample_period = pncc_state->config->framePeriod * 10000000;
      htkHeader.sample_size   = sizeof(float) * pncc_state->config->nCeps;
      htkHeader.sample_kind   = 9; /* USER */
      if (pncc_state->config->cmnOffline || pncc_state->config->cmvnOffline)
        htkHeader.sample_kind += 2048; /* _Z */

      res = WriteHtkHeader(fpOutput, &htkHeader);
    }
    else if (outformat == SPHINX)
    {
      res = WriteSphinxHeader(fpOutput, nCoefs);
    }

    if (res < 0)
    {
      fprintf(stderr, "ERROR: error writing the header to the output file.\n");
      fclose(fpOutput);
      exit(EXIT_FAILURE);
    }

#if defined(USE_DOUBLE)
    outbuf = (float*)fftw_malloc(nCoefs*sizeof(float));
    for (i = 0; i < nCoefs; ++i)
      outbuf[i] = (float)pnccs[i];
#elif defined(USE_FLOAT)
    outbuf = pnccs;
#endif

    nWritten = fwrite(outbuf, sizeof(float), nCoefs, fpOutput);
    fclose(fpOutput);

    if (nWritten != nCoefs)
    {
      fprintf(stderr, "ERROR: error writing the coefficients to the output file.\n");
      exit(EXIT_FAILURE);
    }

#if defined(USE_DOUBLE)
    fftw_free(outbuf);
    fftw_free(pnccs);
#elif defined(USE_FLOAT)
    fftwf_free(pnccs);
#endif
    if (printTiming)
    {
      timersub(&end, &start, &diff);
      double signalDuration = (double)nSamples / (double)sfinfo.samplerate;
      double rtf = (diff.tv_sec + diff.tv_usec / 1000000.0)/signalDuration*100.0;

      fprintf(stdout, "Processing time: %ld s %ld us\n", diff.tv_sec, diff.tv_usec);
      fprintf(stdout, "Signal duration: %.2f s\n", signalDuration);
      fprintf(stdout, "Real-time factor: %.2f%%\n", rtf);
    }
  }
}

int main(int argc, char *argv[])
{
  a3_pncc_config* config;
  a3_pncc_t* pncc_state;
  long int samplerate = 16000;
  bool printTiming = false;
  int c;
  char *input  = NULL;
  char *output = NULL;
  int outformat = SPHINX;
  int option_index = 0;
  bool cmn = false;
  bool cmvn = false;
  long int startTime = LONG_MIN;
  long int endTime   = LONG_MAX;
  a3_real mu = -1.0;
  bool forceRaw = false;

  static struct option long_options[] =
  {
    {"input",       required_argument, 0, 'i'},
    {"output",      required_argument, 0, 'o'},
    {"sample-rate", required_argument, 0, 's'},
    {"start-time",  required_argument, 0, 'a'},
    {"end-time",    required_argument, 0, 'e'},
    {"mu",          required_argument, 0, 'm'},
    {"cmn",         no_argument,       0, 'c'},
    {"cmvn",        no_argument,       0, 'n'},
    {"format",      required_argument, 0, 'f'},
    {"raw",         no_argument,       0, 'r'},
    {"timing",      no_argument,       0, 't'},
    {"version",     no_argument,       0, 'v'},
    {0, 0, 0, 0}
  };

  while (( c = getopt_long (argc, argv, "i:o:s:a:e:m:cnf:rtv", long_options, &option_index)) != -1)
  {
    switch (c)
    {
      case 'i':
        input = optarg;
        break;
      case 'o':
        output = optarg;
        break;
      case 's':
        samplerate = strtol(optarg, NULL, 10);
        if (samplerate == 0L)
        {
          fprintf(stderr, "ERROR: error parsing the --sample-rate argument.\n");
          exit(EXIT_FAILURE);
        }
        if (samplerate != 8000 && samplerate != 16000)
        {
          fprintf(stderr, "ERROR: --sample-rate must be set to 8000 or 16000.\n");
          exit(EXIT_FAILURE);
        }
        break;
      case 'a':
        startTime = strtol(optarg, NULL, 10);
        if (startTime == 0L)
        {
          fprintf(stderr, "ERROR: error parsing the --start-time argument.\n");
          exit(EXIT_FAILURE);
        }
      case 'e':
      endTime = strtol(optarg, NULL, 10);
      if (endTime == 0L)
      {
        fprintf(stderr, "ERROR: error parsing the --end-time argument.\n");
        exit(EXIT_FAILURE);
      }
      case 'c':
        cmn = true;
        break;
      case 'n':
        cmvn = true;
        break;
      case 'f':
        if (strcmp(optarg, "htk") == 0)
        {
          outformat = HTK;
        }
        else if (strcmp(optarg, "sphinx") == 0)
        {
          outformat = SPHINX;
        }
        else
        {
          PrintUsage();
          exit(EXIT_FAILURE);
        }
        break;
      case 'r':
        forceRaw = true;
      case 't':
        printTiming = true;
        break;
      case 'm':
        mu = strtod(optarg, NULL);
        if (mu == 0.0)
        {
          fprintf(stderr, "ERROR: error parsing the --mu argument.\n");
          exit(EXIT_FAILURE);
        }
        break;
      case '?':
        PrintUsage();
        exit(EXIT_FAILURE);
        break;
      case 'v':
        printf("pncc_fe_%s v%s\n", PRECISION_SUFF, VERSION);
        exit(0);
      default:
        abort();
    }
  }

  if (startTime >= endTime)
  {
    fprintf(stderr, "ERROR: start-time cannot be greater than end-time.\n");
    exit(EXIT_FAILURE);
  }

  if (input == NULL || output == NULL)
  {
    PrintUsage();
    exit(EXIT_FAILURE);
  }

  config = NULL;
  if (samplerate == 8000)
    config = a3_pncc_get_def_8khz_config();
  else if (samplerate == 16000)
   config = a3_pncc_get_def_16khz_config();

  config->cmnOffline  = cmn;
  config->cmvnOffline = cmvn;

  if (mu != -1.0)
    config->initial_mean_power = mu;

  if (config == NULL)
  {
    fprintf(stderr, "ERROR: error loading libpncc configuration.\n");
    exit(EXIT_FAILURE);
  }

  pncc_state = a3_pncc_init(config);
  if (pncc_state == NULL)
  {
    fprintf(stderr, "ERROR: error initializing libpncc.\n");
    exit(EXIT_FAILURE);
  }

  ProcessFile(input, output, outformat, pncc_state, printTiming, startTime, endTime, forceRaw);

  a3_real lastmu = a3_get_last_mu(pncc_state);
  printf("lastmu=%f\n", lastmu);

  a3_pncc_free(pncc_state);

  return 0;
}
